<?php

namespace Drupal\mail_box_management\Controller;

use Drupal\Core\Url;
use Drupal\mail_box_management\Plugin\ThemeTemplateLoader;
use Drupal\mail_box_management\Server\Imap\EmailContentManager;
use Drupal\mail_box_management\Server\Imap\ImapConnection;
use Drupal\mail_box_management\Server\Imap\MailBoxStorage;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 *
 */
class MailboxManagementPanel {

  private ThemeTemplateLoader $themeTemplateLoader;

  private ImapConnection $connection;

  public function __construct() {
    $this->themeTemplateLoader = mail_box_management_service('mail_box_management.theme');
    $this->connection = mail_box_management_service('mail_box_management.imap_connector');
  }

  /**
   * @throws \Exception
   */
  public function mailboxManagement(): array|RedirectResponse {
    $content = $this->themeTemplateLoader->getThemeTemplateContent('mail-box-management-panel');
    if(is_null($content)) {
      $uri = Url::fromRoute('mail_box_management.configuration');
      return new RedirectResponse($uri->toString());
    }
    return [
      '#markup' => $content,
    ];
  }

  /**
   * @throws \Exception
   */
  public function mailboxManagementContent(Request $request): JsonResponse {
    $content = json_decode($request->getContent(), TRUE);
    $mailbox = $content['box_name'] ?? NULL;
    $msgno = $content['msgno'] ?? NULL;

    if (empty($mailbox) && empty($msgno)) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'mailbox and msgno are required'], 404);
    }

    $mail = new EmailContentManager($this->connection);
    $mail_found = $mail->getContentWithReferences($mailbox, $msgno);
    $content = $this->themeTemplateLoader->getThemeTemplateContent('mail-box-management-content-view', ['emails' => $mail_found]);
    return new JsonResponse(['status' => TRUE, 'content' => $content->__toString(), 'data' => $mail_found], 200);
  }

  /**
   *
   */
  public function mailboxManagementContentFlag(Request $request): JsonResponse {
    $content = json_decode($request->getContent(), TRUE);
    $mailbox = $content['box_name'] ?? NULL;
    $msgno = $content['msgno'] ?? NULL;
    $flag = $content['flag'] ?? NULL;
    if (empty($mailbox) && empty($msgno) && empty($flag)) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'mailbox and msgno are required'], 404);
    }
    $mail = new EmailContentManager($this->connection);
    return new JsonResponse(['status' => $mail->updateMessageFlags($mailbox, $msgno, strtolower($flag))], 200);
  }

  /**
   *
   */
  public function mailboxManagementContentDelete(Request $request): JsonResponse {
    $content = json_decode($request->getContent(), TRUE);
    $mailbox = $content['box_name'] ?? NULL;
    $msgno = $content['msgno'] ?? NULL;
    if (empty($mailbox) && empty($msgno)) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'mailbox and msgno are required'], 404);
    }
    $mail = new EmailContentManager($this->connection);
    return new JsonResponse(['status' => $mail->deleteEmail($mailbox, $msgno)], 200);
  }

  /**
   * @throws \Exception
   */
  public function mailboxManagementWorkerHandler(Request $request): JsonResponse {
    $content = json_decode($request->getContent(), TRUE);
    $mailbox = $content['mailbox_name'] ?? NULL;
    if (empty($mailbox)) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'mailbox are required'], 404);
    }

    $mail = new EmailContentManager($this->connection);
    $mailbox_contents = $mail->worker($mailbox);
    if (empty($mailbox_contents)) {
      return new JsonResponse([
        'status' => TRUE,
        'list' => [
          $mailbox => [],
        ],
      ], 200);
    }

    $all_contents = [];
    foreach ($mailbox_contents as $mailbox_content) {
      $content = $this->themeTemplateLoader->getThemeTemplateContent('mail-box-management-content-view', ['emails' => $mailbox_content]);
      $mailbox_content = reset($mailbox_content);
      $all_contents[] = [
        trim($mailbox_content['header']['Msgno']) => ['status' => TRUE, 'content' => $content->__toString(), 'data' => $mailbox_content],
      ];
    }
    return new JsonResponse([
      'status' => TRUE,
      'list' => [
        $mailbox => $all_contents,
      ],
    ], 200);
  }

  /**
   * @throws \Exception
   */
  public function mailboxManagementUiRefresh(): JsonResponse {
    $content = $this->themeTemplateLoader->getThemeTemplateContent('mail-box-management-on-refresh');
    if (empty($content->__toString())) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'ui refreshing failed'], 404);
    }
    return new JsonResponse(['status' => TRUE, 'content' => $content->__toString()]);
  }

  /**
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function mailboxManagementComposeMail(Request $request): JsonResponse {

    $content = json_decode($request->getContent(), TRUE);
    if (empty($content['recipient']) || empty($content['subject']) || empty($content['content'])) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'recipient, subject and body content are required'], 404);
    }
    $mail = new EmailContentManager($this->connection);
    $result = $mail->sendMail(
      [
        'recipient' => $content['recipient'],
        'subject' => $content['subject'],
        'message' => $content['content'],
        'attachments' => $content['attachments'],
      ]
    );
    return new JsonResponse(['status' => $result, 'msg' => $mail->message], 200);
  }

  /**
   *
   */
  public function mailboxManagementMailBoxCreation(Request $request): JsonResponse {
    $content = json_decode($request->getContent(), TRUE);
    if (empty($content['title']) || empty($content['type'])) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'title are required'], 404);
    }
    $title = str_replace(' ', '', $content['title']);
    $mailbox = new MailBoxStorage($this->connection);
    if ($content['type'] === 'label') {
      $title = "Labels." . $title;
    }
    return new JsonResponse(['status' => $mailbox->createMilBox($title), 'msg' => 'created successfully'], 200);
  }

  /**
   *
   */
  public function mailboxManagementMailBoxMailTopAction(Request $request): JsonResponse {

    $content = json_decode($request->getContent(), TRUE);
    if (empty($content['type']) || empty($content['msgno']) || empty($content['destination']) || empty($content['source'])) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'destination are required'], 404);
    }
    $mail = new EmailContentManager($this->connection);
    $result = FALSE;
    if ($content['type'] === 'copy') {
      $result = $mail->copyEmails($content['msgno'], $content['source'], $content['destination']);
    }
    elseif ($content['type'] === 'move') {
      $result = $mail->moveEmails($content['msgno'], $content['source'], $content['destination']);
    }

    if ($result) {
      return new JsonResponse(['status' => TRUE, 'msg' => $mail->message], 200);
    }
    return new JsonResponse(['status' => FALSE, 'msg' => $mail->message, 'data' => $content], 404);
  }

  /**
   *
   */
  public function mailboxManagementComposeMailReply(Request $request): JsonResponse {

    $content = json_decode($request->getContent(), TRUE);
    if (empty($content['box_name']) || empty($content['content']) || empty($content['message_id']) || empty($content['subject']) || empty($content['msgno']) || empty($content['recipient'])) {
      return new JsonResponse(['status' => FALSE, 'msg' => 'Reply could`t go through something went wrong'], 404);
    }
    $mail = new EmailContentManager($this->connection);
    $data = $mail->getContent($content['box_name'], $content['msgno']);
    $body_content = $data['body']['content'] ?? NULL;
    $html = "<strong>Reply: " . ($data['header']['Subject'] ?? NULL) . ":</strong><br/><br>";
    $html .= "<blockquote type='cite' style='padding: 0 0.4em; border-left: #1010ff 2px solid; margin: 0;'>{$content['content']}</blockquote>";
    $result = $mail->sendReplyMail([
      'message_id' => $content['message_id'],
      'subject' => "Re:" . $data['header']['Subject'],
      'message' => "<div style='display: block;'>{$body_content}</div>" . $html,
      'recipient' => $content['recipient'],
      'attachments' => $content['attachments'],
    ]);
    return new JsonResponse(['status' => $result, 'msg' => $mail->message], 200);
  }

}
