<?php

namespace Drupal\mail_box_management\Server\Imap;

/**
 * Contains MailBoxStorage class.
 *
 * @file MailBoxStorage.php
 */
/**
 *
 */
class MailBoxStorage {

  private array|bool $mailboxes;

  public function __construct(private readonly ImapConnection $connection) {
    $this->mailboxes = imap_list($this->connection->imap_connection, $this->connection->hostname, '*');
  }

  /**
   *
   */
  public function getMailboxes(): array|bool {
    if (!empty($this->mailboxes)) {
      $new_array = [];
      foreach ($this->mailboxes as $mailbox) {
        $list = explode('}', $mailbox);
        $mailbox_name = end($list);
        if (str_contains($mailbox_name, '.')) {
          $list = explode('.', $mailbox_name);
          $mailbox_name = end($list);
        }
        $new_array[ucfirst(strtolower($mailbox_name))] = $mailbox;
      }
      return $new_array;
    }
    return $this->mailboxes;
  }

  /**
   *
   */
  public function createMilBox(string $mailbox_name): bool {
    $full_name = $this->connection->hostname . "INBOX." . $mailbox_name;
    if (imap_createmailbox($this->connection->imap_connection, imap_utf7_encode($full_name))) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   *
   */
  public function deleteMilBox(string $mailbox_name): bool {
    if (imap_deletemailbox($this->connection->imap_connection, $mailbox_name)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   *
   */
  public function deleteAllMilBoxes(): bool {
    foreach ($this->mailboxes as $mailbox) {
      $this->deleteMilBox($mailbox);
    }
    return TRUE;
  }

  /**
   *
   */
  public function renameMilBox(string $mailbox_name, string $new_name): bool {
    if (imap_renamemailbox($this->connection->imap_connection, $mailbox_name, $new_name)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   *
   */
  public function checkMailboxStatus(string $mailbox_name): array|bool {
    $status = imap_status($this->connection->imap_connection, $mailbox_name, SA_ALL);
    return $status === FALSE ? FALSE : (array) $status;
  }

  /**
   *
   */
  public function getHeaders(string $mailbox_name): array {
    $copy_connection = $this->connection->imap_connection;
    // Open the mailbox.
    $inbox = imap_reopen($copy_connection, $mailbox_name);
    // Check if the mailbox opened successfully.
    $headers = [];
    if ($inbox) {
      // Search for all emails in the "Junks" mailbox.
      $emails = imap_search($copy_connection, 'ALL');

      if ($emails) {
        rsort($emails);
        $emails = count($emails) > 10 ? array_slice($emails, 0, 10) : $emails;
        // Loop through each email and fetch headers and content.
        foreach ($emails as $emailId) {
          // Fetch the email header info.
          $header = imap_headerinfo($copy_connection, $emailId);
          // Check if 'from' and 'to' fields are set and not empty.
          $fromName = !empty($header->from) && isset($header->from[0]->personal) ? $header->from[0]->personal : NULL;
          $fromEmail = !empty($header->from) && isset($header->from[0]->mailbox) ? $header->from[0]->mailbox : NULL;
          $fromHost = !empty($header->from) && isset($header->from[0]->host) ? $header->from[0]->host : NULL;
          $toName = !empty($header->to) && isset($header->to[0]->personal) ? $header->to[0]->personal : NULL;
          $toEmail = !empty($header->to) && isset($header->to[0]->mailbox) ? $header->to[0]->mailbox : NULL;
          $toHost = !empty($header->to) && isset($header->to[0]->host) ? $header->to[0]->host : NULL;

          // Construct full email addresses.
          $fromEmail = $fromEmail ? $fromEmail . '@' . $fromHost : NULL;
          $toEmail = $toEmail ? $toEmail . '@' . $toHost : NULL;

          // Get the subject.
          $subject = $header->subject ?? '';

          // Collect headers.
          $headers[] = [
            'to_mail' => $toEmail,
            'from_mail' => $fromEmail,
            'subject' => $subject,
            'from_name' => $fromName,
            'to_name' => $toName,
            ...((array) $header),
          ];
        }
      }
    }

    // Sort headers by message number if more than 2 headers exist.
    if (count($headers) > 2) {
      usort($headers, function ($a, $b) {
        return intval($b['Msgno']) - intval($a['Msgno']);
      });
    }

    return $headers;
  }

  /**
   *
   */
  public function getTotalRange(string $mailbox_name, int $per_page = 20): array {
    $copy_connection = $this->connection->imap_connection;

    // Open the mailbox.
    $inbox = imap_reopen($copy_connection, $mailbox_name);
    // Check if the mailbox opened successfully.
    if (!$inbox) {
      return [
        'total' => 0,
        'pages' => 0,
      // Return 0 if unable to open mailbox.
      ];
    }

    // Get the total number of messages in the mailbox.
    $totalMessages = imap_num_msg($copy_connection);

    // Calculate the total number of pages needed (round up)
    $pages = ceil($totalMessages / $per_page);
    return [
      'total' => $totalMessages,
      'pages' => $pages,
    ];
  }

  /**
   *
   */
  public function getHeadersForPage(string $mailbox_name, int $page = 1, int $per_page = 20): array {
    $copy_connection = $this->connection->imap_connection;

    // Open the mailbox.
    $inbox = imap_reopen($copy_connection, $mailbox_name);
    if (!$inbox) {
      // Return empty array if unable to open mailbox.
      return [];
    }

    // Get the total number of messages in the mailbox.
    $totalMessages = imap_num_msg($copy_connection);

    // Calculate the starting message number for the given page.
    // The first email on the page.
    $startMessage = ($page - 1) * $per_page + 1;

    // Calculate the ending message number for the given page.
    // Ensure it doesn't exceed total emails.
    $endMessage = min($startMessage + $per_page - 1, $totalMessages);

    // Prepare the email range for imap_search.
    $emailRange = $startMessage . ':' . $endMessage;

    // Search for emails in the given range.
    $emails = imap_search($copy_connection, $emailRange, 'ALL');
    $headers = [];
    if ($emails) {
      // Loop through each email and fetch headers and content.
      foreach ($emails as $emailId) {
        // Fetch the email header info.
        $header = imap_headerinfo($copy_connection, $emailId);
        // Get "From" and "To" emails and names.
        $fromName = $header->from[0]->personal;
        $fromEmail = $header->from[0]->mailbox . '@' . $header->from[0]->host;
        $toName = $header->to[0]->personal;
        $toEmail = $header->to[0]->mailbox . '@' . $header->to[0]->host;
        $subject = $header->subject;
        // $array = (array) $header;
        $headers[] = [
          'to_mail' => $toEmail,
          'from_mail' => $fromEmail,
          'subject' => $subject,
          'from_name' => $fromName,
          'to_name' => $toName,
          ...((array) $header),
        ];
      }
    }
    if (count($headers) > 2) {
      usort($headers, function ($a, $b) {
        return intval($b['Msgno']) - intval($a['Msgno']);
      });
    }

    return $headers;
  }

}
